/**********************************************
  Virtual Pipe Organ for Linux (midi related)
  By Atsushi Ueda <a-ueda@japan-net.ne.jp>
***********************************************/


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#ifdef _LINUX
#include <alsa/asoundlib.h>
#endif

#include "common.h"
#include "main.h"
#include "midi.h"
#include "regs.h"
#include "general_setting.h"
#include "msgbox.h"


extern GENERAL_SETTING general_setting;

#define NOPARAM 99999


#ifdef _WINDOWS
/* 5-9: note messages (channel specific)
 * event data type = snd_seq_ev_note
 */
#define SND_SEQ_EVENT_NOTE              5
#define SND_SEQ_EVENT_NOTEON            6
#define SND_SEQ_EVENT_NOTEOFF           7
#define SND_SEQ_EVENT_KEYPRESS          8
/* 9-10: reserved */

/* 10-19: control messages (channel specific)
 * event data type = snd_seq_ev_ctrl
 */
#define SND_SEQ_EVENT_CONTROLLER        10
#define SND_SEQ_EVENT_PGMCHANGE         11
#define SND_SEQ_EVENT_CHANPRESS         12
#define SND_SEQ_EVENT_PITCHBEND         13      /* from -8192 to 8191 */
#define SND_SEQ_EVENT_CONTROL14         14      /* 14 bit controller value */
#define SND_SEQ_EVENT_NONREGPARAM       15      /* 14 bit NRPN */
#define SND_SEQ_EVENT_REGPARAM          16      /* 14 bit RPN */
/* 18-19: reserved */

#define usleep(n)  printf("%d",n)
#endif


ACT_EVENT act_event[]={
  //action    div ch type                    param   lo hi
  {ACT_CALLREG,0,0,SND_SEQ_EVENT_PGMCHANGE, NOPARAM, 0,127},
  {ACT_CALLREG,0,1,SND_SEQ_EVENT_PGMCHANGE, NOPARAM, 0,127},
  {ACT_CALLREG,0,2,SND_SEQ_EVENT_PGMCHANGE, NOPARAM, 0,127},
  {ACT_CALLREG,0,3,SND_SEQ_EVENT_PGMCHANGE, NOPARAM, 0,127},
  {ACT_VOLUME, 2,0,SND_SEQ_EVENT_CONTROLLER,      1, 0,127},
  {ACT_VOLUME, 2,1,SND_SEQ_EVENT_CONTROLLER,      1, 0,127},
  {ACT_VOLUME, 2,2,SND_SEQ_EVENT_CONTROLLER,      1, 0,127},
  {ACT_VOLUME, 2,3,SND_SEQ_EVENT_CONTROLLER,      1, 0,127}
};

int action_data_num = 8;

static int chorus_level = 750;
static int flg_chorus = 1;
static int flg_open_out = FALSE;
static int key_sts_a[KB_NUM][128];  // actual key status
static int key_sts_c[KB_NUM][128];  // key_status considering coupler selection
static int voice_sts[KB_NUM][STOP_NUM][128]; // voice status
static double pitch = 0.0;

#ifdef _LINUX

static snd_seq_t *seq_handle_out;	//output client
static int out_port[KB_NUM];		//output ports
static snd_seq_t *seq_handle_inp;	//input client
static int inp_port;			//input port

#endif

#ifdef _WINDOWS

#define MIDIHDR_BUFFERSIZE 65536

typedef struct {
  HMIDIOUT  hMidiOut;
  HMIDIIN  hMidiIn;
  MIDIHDR *midihdr;
  int devno;
} PORT;

PORT out_port[KB_NUM];
PORT inp_port;

#endif


void midi_init()
{
  int i,j,k;

  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<128; j++) {
      key_sts_a[i][j] = key_sts_c[i][j] = 0;
      for (k=0; k<STOP_NUM; k++) {
	voice_sts[i][k][j] = 0;
      }
    }
  }

  inp_port = -1;
  for (i=0; i<KB_NUM; i++) {
    out_port[i] = -1;
  }

  seq_handle_inp = NULL;
  seq_handle_out = NULL;

  midi_open_output_client();
}



static int to_midi_ch(int stop)
{
  return (stop >= 9) ? stop+1 : stop;
}


int midi_open_output_client()
{

  int res = snd_seq_open(&seq_handle_out, "default", SND_SEQ_OPEN_OUTPUT, 0);
  if (res<0) {
    fprintf(stderr,"error in midi_open_output_client\n");
    return -1;
  }
  snd_seq_set_client_name(seq_handle_out, "VPO Internal Client (Inaccessible from outside)");
  return 0;
}


/*
  outport 0-3 (65:0-65:3)
 */
int midi_open_output(int port_no, int out_client_id, int outport)
{
#ifdef _LINUX
  int res;

  out_port[port_no] = snd_seq_create_simple_port(seq_handle_out, "Linux Organ_port",
					  SND_SEQ_PORT_CAP_WRITE |
					  SND_SEQ_PORT_CAP_SUBS_READ |
					  SND_SEQ_PORT_CAP_READ,
					  SND_SEQ_PORT_TYPE_MIDI_GENERIC |
					  SND_SEQ_PORT_TYPE_APPLICATION);

  res =  snd_seq_connect_to(seq_handle_out, out_port[port_no], out_client_id, outport);
printf("%d -> %d: %d\n", out_port[port_no], out_client_id, outport);
  return 0;
#else
  UINT rc;
  printf("Opening windows MIDI output device %d...\n", port_no);
  out_port[port_no].devno = port_no;

   if ((rc=midiOutOpen(&out_port[port_no].hMidiOut,
                     out_port[port_no].devno,
                     (DWORD)0L,(DWORD)0L,CALLBACK_FUNCTION)) != MMSYSERR_NOERROR) {
    char msg[300];
    midiOutGetErrorText(rc, msg, sizeof(msg));
    printf("ERROR! %s\n",msg);
  } else {
    out_port[port_no].midihdr = (MIDIHDR*)malloc(sizeof(MIDIHDR));
//    memset(out_port[port_no].midihdr, 0, sizeof(MIDIHDR));
//    out_port[port_no].midihdr->lpData = (char*)malloc(MIDIHDR_BUFFERSIZE);
//    memset(out_port[port_no].midihdr->lpData, 0, MIDIHDR_BUFFERSIZE);
//    out_port[port_no].midihdr->dwBufferLength = MIDIHDR_BUFFERSIZE;
//    midiOutPrepareHeader(out_port[port_no].hMidiOut, out_port[port_no].midihdr,sizeof(MIDIHDR)); 
    //midiOutAddBuffer(out_port[port_no].hMidiIn, out_port[port_no].midihdr,sizeof(MIDIHDR));
    //if(midiOutStart(out_port[port_no].hMidiIn) != MMSYSERR_NOERROR) {
    //  printf("ERROR in opening output midi port !!!\n");
    //}
  }
  return 0;
#endif

}


int midi_close_output()
{
#ifdef _LINUX
  if (flg_open_out == FALSE) return 0;
  snd_seq_close(seq_handle_out);
  flg_open_out = FALSE;
#endif
  return 0;
}



int midi_open_input()
{
#ifdef _LINUX
  int res;
  res = snd_seq_open(&seq_handle_inp, "default", SND_SEQ_OPEN_INPUT, 0);
  snd_seq_set_client_name(seq_handle_inp, "Linux Organ");

  inp_port = snd_seq_create_simple_port(seq_handle_inp, "Linux Organ Input Port",
                                       SND_SEQ_PORT_CAP_WRITE |
                                       SND_SEQ_PORT_CAP_SUBS_WRITE |
                                       SND_SEQ_PORT_CAP_WRITE,
                                       SND_SEQ_PORT_TYPE_MIDI_GENERIC |
                                       SND_SEQ_PORT_TYPE_APPLICATION);
  //printf("%d\n", p->my_port);

//  res=snd_seq_connect_from(p->seq_handle, p->my_port, 64, 0);
#else
  UINT rc;
  printf("Opening windows MIDI input device...\n");
  inp_port.devno = general_get_input_dev();
  if ((rc=midiInOpen(&inp_port.hMidiIn,
                     inp_port.devno,
                     midi_in_proc,(DWORD)0L,CALLBACK_FUNCTION)) != MMSYSERR_NOERROR) {
    char msg[300];
    midiInGetErrorText(rc, msg, sizeof(msg));
    printf("ERROR! %s\n",msg);
  } else {
    inp_port.midihdr = (MIDIHDR*)malloc(sizeof(MIDIHDR));
    memset(inp_port.midihdr, 0, sizeof(MIDIHDR));
    inp_port.midihdr->lpData = (char*)malloc(MIDIHDR_BUFFERSIZE);
    memset(inp_port.midihdr->lpData, 0, MIDIHDR_BUFFERSIZE);
    inp_port.midihdr->dwBufferLength = MIDIHDR_BUFFERSIZE;
    midiInPrepareHeader(inp_port.hMidiIn, inp_port.midihdr,sizeof(MIDIHDR)); 
    midiInAddBuffer(inp_port.hMidiIn, inp_port.midihdr,sizeof(MIDIHDR));
    if(midiInStart(inp_port.hMidiIn) != MMSYSERR_NOERROR) {
      printf("ERROR in opening input midi port !!!\n");
    }
  }
#endif
  return 0;
}


#ifdef _LINUX
static void dmp_event(snd_seq_event_t *e)
{
  int f=0;

  printf("dest=%d:%d\t", e->dest.client, e->dest.port);
  printf("type=%2d ", e->type);
  switch (e->type) {
  case SND_SEQ_EVENT_NOTEON:
    printf("noteon\t");
    printf("ch=%d ", e->data.note.channel+1);
    printf("note=%d ",e->data.note.note);
    printf("vel=%d\n", e->data.note.velocity);
    break;
  case SND_SEQ_EVENT_NOTEOFF:
    printf("noteoff\t");
    printf("ch=%d ", e->data.note.channel+1);
    printf("note=%d ",e->data.note.note);
    printf("vel=%d\n", e->data.note.velocity);
    break;
  case SND_SEQ_EVENT_CONTROLLER:
    if (!(f++)) printf("ctrl\t");
  case SND_SEQ_EVENT_PGMCHANGE:
    if (!(f++)) printf("pgmchg\t");
  case SND_SEQ_EVENT_CHANPRESS:
    if (!(f++)) printf("chanpress\t");
  case SND_SEQ_EVENT_PITCHBEND:
    if (!(f++)) printf("pitchbend\t");
  case SND_SEQ_EVENT_CONTROL14:
    if (!(f++)) printf("control14\t");
  case SND_SEQ_EVENT_NONREGPARAM:
    if (!(f++)) printf("nonregparam\t");
  case SND_SEQ_EVENT_REGPARAM:
    if (!(f++)) printf("regparam\t");
    printf("ch=%d ", e->data.control.channel+1);
    printf("param=%d ", e->data.control.param);
    printf("value=%d\n", e->data.control.value);
    break;
  default:
    printf("\n");
  }
}


static void  send_event(int kb_no, snd_seq_event_t *ev, int do_flush)
{
  int res;

  if (flg_open_out == FALSE) {
    printf("send_event() is called when flg_open_out==FALSE. The following event could not be sent.\n");
    dmp_event(ev);
    return;
  }

  snd_seq_ev_set_source(ev, out_port[kb_no]);
  snd_seq_ev_set_subs(ev);
  snd_seq_ev_set_direct(ev);
  res = snd_seq_event_output(seq_handle_out, ev);


  if (res < 0) {
	printf("error in send_event(). The following event could not be sent.\n");
	dmp_event(ev);
    return;
  }

  if (do_flush)
    snd_seq_drain_output(seq_handle_out);
}

#endif


static void note_on(int port_no, int chan_no, int note, int vel)
{
#ifdef _LINUX
  snd_seq_event_t ev;
  snd_seq_ev_set_noteon(&ev, chan_no, note, vel);
  send_event(port_no, &ev, 1);
#else
  midiOutShortMsg(out_port[port_no].hMidiOut, (0x90+chan_no) | (note << 8) | (vel << 16)); 
#endif
}

static void note_off(int port_no, int chan_no, int note, int vel)
{
#ifdef _LINUX
  snd_seq_event_t ev;
  snd_seq_ev_set_noteoff(&ev, chan_no, note, vel);
  send_event(port_no, &ev, 1);
#else
  midiOutShortMsg(out_port[port_no].hMidiOut, (0x80+chan_no) | (note << 8) | (0 << 16)); 
#endif
}


static void program(int port_no, int chan_no,  int bank, int preset)
{
#ifdef _LINUX
  snd_seq_event_t ev;
  snd_seq_ev_set_controller(&ev, chan_no, 0, bank);
  send_event(port_no, &ev, 0);

  snd_seq_ev_set_pgmchange(&ev, chan_no, preset);
  send_event(port_no, &ev, 0);
#else
  midiOutShortMsg(out_port[port_no].hMidiOut, (0xb0+chan_no) | (0x00 << 8) | (bank  << 16)); 
  midiOutShortMsg(out_port[port_no].hMidiOut, (0xb0+chan_no) | (0x20 << 8) | (0  << 16)); 
  midiOutShortMsg(out_port[port_no].hMidiOut, (0xc0+chan_no) | (preset << 8) | (0 << 16)); 
#endif
}

static void control(int port_no, int chan_no, int type, int val)
{
#ifdef _LINUX
  snd_seq_event_t ev;
  //printf("port_no=%d chan_no=%d type=%d val=%d\n", port_no,chan_no,type,val);
  //return;
  snd_seq_ev_set_controller(&ev, chan_no, type, val);
  send_event(port_no, &ev, 1);
#else
  midiOutShortMsg(out_port[port_no].hMidiOut, (0xb0+chan_no) | (type << 8) | (val << 16)); 
#endif
}

static void pitchbend(int port_no, int chan_no, int val)
{
#ifdef _LINUX
  snd_seq_event_t ev;
  snd_seq_ev_set_pitchbend(&ev,chan_no,val-8192);
  send_event(port_no, &ev, 1);
#else
//  midiOutShortMsg(out_port[port_no].hMidiOut, (0xe0+chan_no) | (((val&0x3f8)>>7)*0 << 8) | ((val&0x7f) << 8));
  int val_h = (val & 0x03f80)>>7;
  int val_l = val&0x7f;

  midiOutShortMsg(out_port[port_no].hMidiOut, (0xe0+chan_no) | (val_l << 8) | (val_h << 16)); 
  printf("   pbend H=%02X L=%02X  val=%04X (%d)\n", val_h, val_l, val,val);
#endif

}


void midi_set_preset()
{
  int i,j;
  for (i=0; i<KB_NUM; i++) {
    int bank, preset;
    for (j=0; j<15; j++) {
      regs_get_preset(i,j,&bank,&preset);
      if (bank>=0 && bank<=127 && preset>=0 && preset<=127) {
	program(i, to_midi_ch(j), bank, preset); 
	//printf("[%d,%d,%d,%d]",  i,   to_midi_ch(j), bank, preset  ); 
      }
    }
  }
}


static void midi_action_event(int channel, int type, int param, int value) 
{
  extern GtkWidget *main_window;
  int i, ch;
  for (i=0; i<action_data_num; i++) {
    if (act_event[i].event_type == type) {
      if (act_event[i].event_param == param ||
	  act_event[i].event_param==NOPARAM) {
	if (act_event[i].input_ch == channel) {
	  switch (act_event[i].action) {
	  case ACT_CALLREG:
	    gdk_threads_enter();
	    combination_btn(NULL, (gpointer)(1+value - act_event[i].range_low));
	    for (i=0; i<1; i++) {
	      gtk_widget_queue_draw (main_window); 
	    }
	    gdk_threads_leave();
	    break;
	  case ACT_VOLUME:
	    for (ch=0; ch<16; ch++) {
	      int volume;
	      const int min_vol = 12;
	      const int max_vol = 127;
	      int low = act_event[i].range_low;
	      int high = act_event[i].range_high;
	      if (value < low) { 
		value = low;
	      }
	      if (value > high) {
		value = high;
	      }
	      volume = min_vol + (max_vol-min_vol)*(value-low)/(high-low);
	      control(act_event[i].action_div, 
		      ch, 
		      7, 
		      volume);
	    printf("volume %d\n", volume);
	    }
	    break;
	  }
	}
      }
    }
  }
}



void midi_key_input(int *ch, int *note, int *vel)
{
#ifdef _LINUX
  for (;;) {
    int res;
    int type;
    snd_seq_event_t *ev;
    res = snd_seq_event_input(seq_handle_inp, &ev);
    type = ev->type;

    if (ev->type!=36){

      dmp_event(ev);
      midi_action_event(ev->data.control.channel,ev->type,ev->data.control.param,ev->data.control.value);
    }

    if (type != SND_SEQ_EVENT_NOTEON && type!=SND_SEQ_EVENT_NOTEOFF) continue;

    *ch = ev->data.note.channel;
    *note = ev->data.note.note;
    *vel = ev->data.note.velocity;
    
    if (type == SND_SEQ_EVENT_NOTEOFF) {
      *vel = 0;
    }

    return;

  }
#endif
}


void midi_set_stop_volume()
{
  int i,j;
  return;
  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<15; j++) {
      if (regs_get_current_stop_select(i,j)) {
	control(i, to_midi_ch(j), 7, 80);
	//printf("div=%d  stop=%d  on\n", i, j);
      } else {
	control(i, to_midi_ch(j), 7, 80);
	//printf("div=%d  stop=%d  off\n", i, j);
      }
    }
  }
}

void midi_enable_chorus(int flg)
{
  int i,j;
  flg_chorus = flg;
  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<15; j++) {
      int level = (int)(8192 + (rand() % chorus_level - chorus_level/2) + (pitch*4096));
      if (level<-0) {
        level=-0;
      } else if (level>16383) {
        level=16383;
      }
      pitchbend(i,to_midi_ch(j),  flg ? level : (int)(8192+(pitch*4096)));
      
    }
  }
}

int midi_get_chorus_level()
{
  return chorus_level;
}

void midi_set_chorus_level(int level)
{
  chorus_level = level;
  if (flg_open_out == TRUE) {
    midi_enable_chorus(flg_chorus);
  }
}

double midi_get_pitch_level()
{
  return pitch;
}

void midi_set_pitch_level(double level)
{
  pitch = level;
  midi_enable_chorus(flg_chorus);
}




void midi_enable_tremulant(int div, int flg)
{
  int i;
  for (i=0; i<15; i++) {
    control(div, to_midi_ch(i), 1, flg?30:0);
  }
}


void midi_set_one_key_cp(int note)
{
  int i, j;
  for (i=0; i<KB_NUM; i++) {
    key_sts_c[i][note]=0;
  }
  for (i=0; i<KB_NUM; i++) {
    key_sts_c[i][note] |= key_sts_a[i][note];
    for (j=0; j<KB_NUM; j++) {
      if (regs_get_current_coupler_state(i,j)) {
	key_sts_c[j][note] |= key_sts_a[i][note];
	//	printf("add %d to %d  note=%d  %d\n", i,j,note, key_sts_c[j][note]);
      }
    }
  }
}


void midi_set_one_voice_sts(int note)
{
  int i,j,sts;
  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<STOP_NUM; j++) {
      sts = key_sts_c[i][note] && regs_get_current_stop_select(i,j);
      if (sts != voice_sts[i][j][note]) {
	voice_sts[i][j][note] = sts;
	if (sts) {
	  note_on(i,
		  to_midi_ch(j),
		  note,
		  regs_get_current_stop_velocity(i, j));
	} else {
	  note_off(i,
		   to_midi_ch(j),
		   note,0);
	}
      }
    }
  }
}


void midi_set_all_voice_sts()
{
  int i;
  for (i=0; i<128; i++) {
    midi_set_one_key_cp(i);
    midi_set_one_voice_sts(i);
  }
}


void midi_main_loop(void *ptr)
{

  int i;
  for (;;) {
    int ch, note0, note, vel;
    midi_key_input(&ch, &note0, &vel);

    for (i=0; i<KB_NUM; i++) {
      if (general_get_div_channel(i) == ch) {
	note = note0 + general_get_div_transpose(i);
	if (note<0) note=0;
	if (note>127) note=127;
	key_sts_a[i][note] = vel ? 1 : 0;
	midi_set_one_key_cp(note);
	midi_set_one_voice_sts(note);
	//break;
      }
    }
  }
}

#ifdef _WINDOWS

void CALLBACK midi_in_proc(HMIDIIN hMidiIn, UINT wMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2)
{
  int ch;
  int type;
  int param;
  int value;

  if(wMsg==MIM_DATA){
    type=0;
    ch=dwParam1 & 0x0000000f;
    param=(dwParam1>>8) & 0xff;
    value=(dwParam1>>16) & 0xff;
    switch((int)(dwParam1&0x000000f0)) {
      
    case 0x80:
      type = SND_SEQ_EVENT_NOTEOFF;
      value=0;
      break;
      
    case 0x90:
      if (value == 0)
        type = SND_SEQ_EVENT_NOTEOFF;
      else
        type = SND_SEQ_EVENT_NOTEON;
      break;
    case 0xc0:
      type = SND_SEQ_EVENT_PGMCHANGE;
      break;
    case 0xe0:
      type = SND_SEQ_EVENT_PITCHBEND;
      printf("%02x %02x\n", param,value);
      break;
    case 0xf0:
      type = SND_SEQ_EVENT_CONTROLLER;
      
    default:
      break;
    }
  }
  if (ch==8) return;
  printf("ch=%d type=%d param=%d value=%d\n", ch,type,param,value);
  if (type!=SND_SEQ_EVENT_NOTEON && type!=SND_SEQ_EVENT_NOTEOFF) {
    return;
  }
  {
    int note;
    int i;
    for (i=0; i<KB_NUM; i++) {
      if (general_get_div_channel(i) == ch) {
        note = param + general_get_div_transpose(i);
        if (note<0) note=0;
        if (note>127) note=127;
        key_sts_a[i][note] = value ? 1 : 0;
        midi_set_one_key_cp(note);
        midi_set_one_voice_sts(note);
      }
    }
  }
}

#endif

int midi_get_timidity_client_id()
{
#ifdef __ASOUNDLIB_H
  int rc = -1;
  int res;
  int i,j;
  snd_seq_t *seq_handle;

  char *name;
  snd_seq_port_info_t *info;
  snd_seq_port_info_alloca(&info);

  res = snd_seq_open(&seq_handle, "hw", SND_SEQ_OPEN_OUTPUT, 0);
  if (res) return rc;

  for (i=0; i<255; i++) {
    for (j=0; j<8; j++) {
      res = snd_seq_get_any_port_info(seq_handle, i,j,  info);
      if (res) continue;
      name = (char*)snd_seq_port_info_get_name(info);
      if (name) {
	if (!strcmp(name, "TiMidity port 0")) {
	  rc = i;
	}
      }
    }
  }
  snd_seq_close(seq_handle);

  return rc;
#else
  return 0;
#endif
}



#define gs general_setting


char* midi_make_timidity_cmdline(char *sffile_name)
{
  static char cmdline[1000];
  int len = 0;
  
  len += sprintf(cmdline+len, "%s ", gs.timi_location);

#ifdef _LINUX
  len += sprintf(cmdline+len, "-iA ");
#endif
#ifdef _WINDOWS
  len += sprintf(cmdline+len, "-iW ");
#endif

  len += sprintf(cmdline+len, "-C%d ", gs.timi_ctlrate);
  len += sprintf(cmdline+len, "-EFreverb=%d,%d ", gs.timi_reverb?1:0,gs.timi_reverb);
  len += sprintf(cmdline+len, "-EFchorus=%d,%d ", gs.timi_chorus?1:0,gs.timi_chorus);
  len += sprintf(cmdline+len, "-p%d ", gs.timi_voices);
  len += sprintf(cmdline+len, "%s ", gs.timi_options);
#ifdef _LINUX
  //len += sprintf(cmdline+len, "-x'soundfont %s order=0' ", sffile_name );
  len += sprintf(cmdline+len, "--config-file=%stmp.cfg ", general_get_rc_dir());
#endif
#ifdef _WINDOWS
  len += sprintf(cmdline+len, "-x\x22soundfont %s order=0\x22 ", sffile_name );
#endif
  if (gs.eca_enable) {
    len += sprintf(cmdline+len, "-Or -o - ");
    len += sprintf(cmdline+len, "| %s ", gs.eca_location);
    len += sprintf(cmdline+len, "%s ", gs.eca_options);
  } else {
  }
#ifdef _LINUX
  len += sprintf(cmdline+len, "&");
#endif
#ifdef _WINDOWS
  len += sprintf(cmdline+len, "< a.txt");
  strrep(cmdline, "\\","\\\\");
#endif
  
  return cmdline;
}


#ifdef _WINDOWS

static int timi_running = 0;
static char* cmd_timidity;

void midi_launch_timidity()
{
  timi_running = 1;
  system(cmd_timidity);
  timi_running = 0;
}

#endif



int midi_launch_timidity_and_connect(char *sffile_name)
{
#ifdef _LINUX
  int client_id = -1;
  int i;
  char *cmdline;

  midi_close_output();

  if (midi_get_timidity_client_id()>=0) {
    char cmdline[200];
    sprintf(cmdline, "killall %s", exename(gs.timi_location));
    puts(cmdline);
    system(cmdline);

    for (i=0; i<20; i++) {
      client_id = midi_get_timidity_client_id();
      if (client_id < 0) break;
      usleep (100000);
    }
    usleep (100000);
  }


  if (client_id >= 0) {
     msgbox("Existing Timidity termination failed.","error","OK",0);
     return FALSE;
  }

  cmdline = midi_make_timidity_cmdline(sffile_name);
  
  printf("<<<<%s\n>>>>",cmdline);

  system(cmdline);

  for (i=0; i<50; i++) {
    client_id = midi_get_timidity_client_id();
    if (client_id > 0) break;
    usleep (100000);
  }

  if (client_id < 0) return FALSE;

  printf("timidity client_id = %d\n", client_id);

  midi_open_output_client();

  for (i=0; i<KB_NUM; i++) {
    midi_open_output(i, client_id, i);
  }
  flg_open_out = TRUE;

  midi_enable_chorus(flg_chorus);
  usleep(100000);

  return client_id;

#endif

#ifdef _WINDOWS

  printf("%d\n",KILL_PROC_BY_NAME("timidity.exe"));
  
  cmd_timidity = midi_make_timidity_cmdline(sffile_name);
  puts(cmd_timidity);
  _beginthread(midi_launch_timidity, 10000, NULL);
  Sleep(2000);

#endif

}


